#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>

#include <lber.h>
#include <ldap.h>

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef	HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef	HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include <strfunc.h>

#include "lgba.h"

extern char *lgba_i_user;
extern LDAP *ld;

int cmd_help(slist *);
int cmd_user(slist *);
int cmd_lang(slist *);
int cmd_node(slist *);
int cmd_help(slist *);
int cmd_create(slist *);
int cmd_cd(slist *);
int cmd_ls(slist *);
int cmd_load(slist *);
int cmd_show(slist *);
int cmd_class(slist *);
int cmd_rm(slist *);
int cmd_arm(slist *);
int cmd_crm(slist *);
int cmd_set(slist *);
int cmd_add(slist *);
int cmd_ov(slist *);
int cmd_file(slist *);
int cmd_cec(slist *);
int cmd_fec(slist *);
int cmd_vp(slist *);
int cmd_exit(slist *);

typedef struct {
	char *cmd;
	int (*func)(slist *);
	char *dsc;
} CMDS;


CMDS cmds[] = {
{ "?", cmd_help, "Display this page" },
{ "help", cmd_help, "Display this page" },
{ "user", cmd_user, "Set active user" },
{ "lang", cmd_lang, "Set active language" },
{ "node", cmd_node, "Change current node" },
{ "cd", cmd_cd, "Change current dn" },
{ "ls", cmd_ls, "List all objects ('ls -h' shows other options)" },
{ "create", cmd_create, "Create object" },
{ "load", cmd_load, "Load object silently" },
{ "show", cmd_show, "Load and show object" },
{ "class", cmd_class, "Add class to the current object" },
{ "rm",	cmd_rm, "Remove object" },
{ "arm", cmd_arm, "Remove particular attribute" },
{ "crm", cmd_crm, "Remove object from the specified class" },
{ "set", cmd_set, "Set a value to an attribute" },
{ "add", cmd_add, "Add a value to an attribute" },
{ "vp", cmd_vp, "Verify password" },
{ "override", cmd_ov, "Set 'override' flag for friendly output (default is 1)" },
{ "file", cmd_file, "Create formatted files" },
{ "cec", cmd_cec, "Clear error code" },
{ "fec", cmd_fec, "Freeze error code" },
{ "exit", cmd_exit, "Exit lgba" },
{ NULL, NULL, NULL }
};

char *flt;
char *fmt;
char *exec_format;

int l_exec(const char *s);

int f_cnt(int);
int f_fmt(int);

int rm_complex(char *attr, int quoted);

#ifdef	HAVE_LIBREADLINE
char **lgba_completion(char *text, int start, int end);
char *objects_generator(char *, int);
char *attributes_generator(char *, int);
char *commands_generator(char *, int);
#endif

int frozen=0, fcode=0;

int main(int ac, char **av) {
    char *s;
    char *prompt=NULL;
    int promlen=32;
    int t;
	char **cmds=NULL;
	sbuf *sb;

    prompt=(char *)sf_malloc(promlen);
    strcpy(prompt, "LGBA>");
   
	if(lgba_init(NULL)) {
		fprintf(stderr, "-ERR: LGBA init failed.\n");
		return -1;
	};
	if(lgba_open()) {
		fprintf(stderr, "-ERR: Connection to LDAP server failed.\n");
		return -1;
	};
	setgid(getgid());

	if(ac > 1) {
		slist *sl=sinit();
		slist *sl2;
		int i;
		for(i=1; i<ac; i++)
			sadd(sl, av[i]);
		sl2 = split(sjoin(sl, " "), ";", 0);
		cmds=mkarray(sl2, 0);
		sfree(sl2);
		sfree(sl);
	};

	if(cmds) {

		for(; *cmds; cmds++) {
			if(eq(*cmds, "-")) break;
			l_exec(*cmds);
		};

		if(!*cmds) goto quit;

		cmds++;
	};

#ifdef	HAVE_LIBREADLINE
	/* Initialize readline */
	rl_readline_name="LGBA";
	rl_completion_entry_function = (Function *)objects_generator;
	rl_attempted_completion_function = lgba_completion;
#endif

here:

	if(isatty(0) && isatty(1)) {

#ifdef	HAVE_LIBREADLINE
	  while((s=readline(prompt))) {
#else
	  fprintf(stderr, "%s", prompt);
	  sb = sbuf_init();
	  while((sbuf_zero(sb) == 0) && (s=sbuf_fgets(sb, stdin))) {
#endif

		if(*s) {
#ifdef	HAVE_LIBREADLINE
			add_history(s);
#endif
			l_exec(s);
		}
#ifdef	HAVE_LIBREADLINE
		free(s);
#endif
		if(!(s=lgba_objectname())) {
			strcpy(prompt, "LGBA>");
		} else {
			t=strlen(s);
			while((t + 10) >= promlen)
				prompt=(char *)sf_realloc(prompt, promlen<<=1);
			sprintf(prompt, "LGBA:%s>", s);
		};
#ifndef	HAVE_LIBREADLINE
	  fprintf(stderr, "%s", prompt);
#endif
	  };
    	  printf("\n");
	} else {
	    sb = sbuf_init();
		while((sbuf_zero(sb) == 0) && (s=sbuf_fgets(sb, stdin)))
			l_exec(s);
	};

	if(cmds) {

		while(*cmds) {
			if(eq(*cmds, "-")) break;
			l_exec(*cmds);
			cmds++;
		};

		if(*cmds++) goto here;

	};

quit:

	lgba_close();
    return(frozen?fcode:lgba_errno);
};

int
l_exec(const char *func) {
	int slen;
	static slist *sl=NULL;
	CMDS *cmd;
	char *p;

	if(!func || !*func) return LGBA_NOERR;

	if(!sl) sl=sinit();
	sclear(sl);
	
	splitquotable(sl, func);
	if(!sl->count) return LGBA_NOERR;


	/* Make command lower-case. */
	slen=sl->lens[0];
	for( p=sl->list[0]; *p; p++ )
		*p = tolower(*p);

	for( cmd=cmds; cmd->cmd; cmd++) {
		if(!strncmp(sl->list[0], cmd->cmd, (slen>2)?slen:2)) {
			sdel(sl, 0);
			return cmd->func(sl);
		};
	};

	fprintf(stderr, "-ERR: Unknown command.\n");
	return lgba_errno=LGBA_COMMAND;
};

int f_cnt(int z) {
	if(!z)
		fprintf(stderr, "!NOTE: No objects found.\n");
	return z;
};

int f_fmt(int z) {
	format(fmt, 1);
	return 0;
};


int 
rm_complex(char *attr, int quoted) { 
    char *af, *as;

    if(!attr)
		return lgba_errno=LGBA_COMMAND;

    if(quoted || !(as=strchr(attr, '=')))
		return rmattr(attr, NULL);

    // Splitting attr.
    af=attr;
    *as++=0;
    
	return rmattr(attr, as);
}

#ifdef	HAVE_LIBREADLINE
char **
lgba_completion(char *text, int start, int end) {
	char **matches=NULL;

	if(start == 0)
		matches = completion_matches(text, commands_generator);
	else {
		if(lgba_objectname()) {
			char *p=text;
			if(*p == '/' || *p == '~' || *p == '.')
				matches =
			completion_matches(text, filename_completion_function);
			else
			matches = completion_matches(text, attributes_generator);
		} else {
			matches = completion_matches(text, objects_generator);
		};
	};

	return matches;
};
#endif

char *
commands_generator(char *text, int state) {
	static int idx, len;
	char *name;

	if (!state) {
		idx = 0;
		len = strlen (text);
	}

	while((name = cmds[idx].cmd)) {
		idx++;
		if(!strncmp(name, text, len)) return sf_strdup(name);
	}

 	return NULL;
};

int cmd_help(slist *sl) {
		int i;
		for(i = 0; i < sizeof(cmds) / sizeof(CMDS) - 1; i++)
			fprintf(stderr, "%-10s:\t%s\n", cmds[i].cmd, cmds[i].dsc);
		return LGBA_NOERR;
};

int cmd_user(slist *sl) {
		if(sl->count == 0) {
			fprintf(stderr, "$INFO: User is '%s'.\n", lgba_i_user);
			return LGBA_NOERR;
		}
		if(lgba_setuser(sl->list[0])) {
			lgba_perror("-ERR");
			return lgba_errno;
		};
		fprintf(stderr, "$INFO: User set to '%s'.\n", lgba_i_user);
		return LGBA_NOERR;
};

int
_lgba_srt(const char *a, const char *b) {
	register int va=atoi(a);
	register int vb=atoi(b);
	if(va != vb) return va-vb;
#ifdef  HAVE_STRCOLL
	return strcoll(a, b);
#else
	return strcmp(a, b);
#endif
};

int
_lgba_rev_srt(const char *b, const char *a) {
	register int va=atoi(a);
	register int vb=atoi(b);
	if(va != vb) return va-vb;
#ifdef  HAVE_STRCOLL
	return strcoll(a, b);
#else
	return strcmp(a, b);
#endif
};


int
cmd_ls(slist *sl) {
	int ret;
	int c;
	int ac;
	char **av;
	int flags=0;
	char *sortby=NULL;
	int countonly = 0;
	int (*cmp)(const char *, const char *)=_lgba_srt;

#define	FLT	"(|(!(objectclass=*))(objectclass=*))"
#define	FMT	"${~}:${objectclass}\t(${cn})\\n"

	flt=NULL;
	fmt=NULL;
	exec_format=NULL;

	sins(sl, "ls", 0);

	optind = 1;
	ac=sl->count;
	av=sl->list;

	while((c = getopt(ac, av, "Rcrths:e:f:")) != -1)
	switch(c) {
		case 'c':
			countonly=1;
		case 'f':
			if(optarg) {
				if(!fmt)
					fmt = sf_strdup(optarg);
			} else {
				if(fmt)
					free(fmt);
				fmt = NULL;
			};
			break;
		case 'R':	/* replay search */
			flags |= 2;
			break;
		case 'r':
			cmp = _lgba_rev_srt;
			break;
		case 't':	/* subtree search */
			flags |= 1;
			break;
		case 's':
			sortby = optarg;
			break;
		case 'e':
			exec_format = optarg;
			break;
		case 'h':
			fprintf(stderr, "Usage: ls [<options>] [<LDAP_filter> [<format> ...]]\n");
			fprintf(stderr, "Options are:\n");
			fprintf(stderr, "\t-h                 This page\n");
			fprintf(stderr, "\t-c                 Count only\n");
			fprintf(stderr, "\t-f <format_string> Format string\n");
			fprintf(stderr, "\t-R:                Redisplay search results\n");
			fprintf(stderr, "\t-r                 Reverse sort\n");
			fprintf(stderr, "\t-s <attribute>     Sort by specified attribute\n");
			fprintf(stderr, "\t-t                 Subtree search\n");
			fprintf(stderr, "\t-e <cmd_format>    Invoke subshell command for every found entry\n");
			return LGBA_NOERR;
		default:
			fprintf(stderr, "-ERR: Unknown option: -%c\n", c);
			return LGBA_COMMAND;
	};

	ac-=optind;
	av+=optind;

	if(ac == 0) {
		if(!flt) flt=sf_strdup(FLT);
		if(!fmt) fmt=sf_strdup(FMT);
	} else if(ac == 1) {
		if(!flt) {
			flt = sf_strdup(*av);
			if(!fmt)
				fmt = sf_strdup(FMT);
		} else if(!fmt) {
			fmt = sf_strdup(join(av, " "));
		};
	} else {
		if(!flt)
			flt = sf_strdup(*av++);
		if(!fmt) {
			if(ac) fmt = sf_strdup(join(av, " "));
			else fmt = sf_strdup(FMT);
		};
	};

	/* Save changes, if necessary */
	if(lgba_save(1))
		lgba_perror("-ERR: Error saving previous changes");

	av = NULL;
	if(countonly) {
		slist *sl=sinit();
		sadd(sl, "ObjectClass");
		sadd(sl, lgba_ObjectAttribute());
		av = mkarray(sl, 0);
		sfree(sl);
	};

	ret=lgba_errno;
	lgba_errno=0;

	if(countonly)
		c = lgba_search(NULL, flt, av, NULL, NULL, f_cnt, NULL, flags);
	else
		lgba_search(NULL, flt, av, sortby, cmp, f_cnt, f_fmt, flags);

	if(av)
		free_values(av);

	if(lgba_errno) {
		lgba_perror("-ERR: LDAP Search error");
		ret=lgba_errno;
	} else {
		lgba_errno=ret;
		if(countonly)
			printf("%d\n", c);
	};

	free(flt);
	free(fmt);
	return ret;
}

int cmd_lang(slist *sl) {
	if(sl->count == 0) {
		fprintf(stderr, "$Usage: lang <language>\n");
		return LGBA_NOERR;
	}
	if(!lgba_setlang(sl->list[0]))
		return lgba_errno=LGBA_SETLANG;
	return LGBA_NOERR;
};

int cmd_node(slist *sl) {
	char const *s;
	if(sl->count == 0) {
		fprintf(stderr, "Usage: chnode <name>\n");
		if((s = lgba_current_node()))
			fprintf(stderr, "Current node: %s\n", s);
		fprintf(stderr, "Available nodes: %s\n",
			sjoin(lgba_nodes(), ", "));
		return LGBA_NOERR;
	}

	s=sjoin(sl, " ");
	if(lgba_chnode(s))
		lgba_perror("-ERR");
	else
		fprintf(stderr, "+OK: Node changed.\n");
	return lgba_errno;
};

int cmd_cd(slist *sl) {
	char *s;
	s=sjoin(sl, " ");
	if(lgba_cd(s))
		lgba_perror("-ERR");
	else
		fprintf(stderr, "+OK: Current DN changed to '%s'.\n",
			lgba_current_dn());
	return lgba_errno;
};

int cmd_create(slist *sl) {
	if(sl->count == 0) {
		fprintf(stderr, "$INFO: Usage: create <object> [ <objectclass> ]\n");
		return lgba_errno=LGBA_COMMAND;
	}
	if(create(sl->list[0], sl->list[1])) {
		lgba_perror("-ERR");
		return lgba_errno=LGBA_COMMAND;
	}
	fprintf(stderr, "+OK: %s (%s) created.\n",
		lgba_objectname(), lgba_objectdn());
	return LGBA_NOERR;
};

int cmd_load(slist *sl) {
	if(sl->count == 0) {
		fprintf(stderr, "$INFO: Usage: load <object>.\n");
		return LGBA_NOERR;
	};
	if(lgba_save(1))
		lgba_perror("-ERR: Error saving previous entry");
	if(load((const char *)sl->list[0]))
		lgba_perror("-ERR");
	else
		fprintf(stderr, "+OK: Object loaded.\n");
	return lgba_errno;
};

int cmd_show(slist *sl) {
	if(sl->count > 0) {
		if(lgba_save(1))
		lgba_perror("-ERR: Error saving previous entry");
		if(load(sl->list[0])) {
			lgba_perror("-ERR");
			return lgba_errno;
		};
	};
	if(!lgba_objectdn()) {
		fprintf(stderr, "-ERR: Object not loaded. Use load <object>.\n");
		return lgba_errno=LGBA_COMMAND;
	};
	if(sl->count == 0)
		if(load(lgba_objectname())) {
			fprintf(stderr, "-ERR: Failed to reload object.\n");
			return lgba_errno=LGBA_LDAP;
		};

{
    int i, m=0;
    char *p;
    char *ts;
	char **values, **vp;
	slist *slt;

	if(!(values=lgba_objectattributes())) {
		fprintf(stderr, "-ERR: Failed get object attributes.\n");
		return lgba_errno=LGBA_LDAP;
	};
	

    // Searching for the maximum attribute length.
	for(vp = values; *vp; vp++) {
		if((i=strlen(*vp)) > m) m=i;
		if((i=strlen(lgba_map_friendly(*vp))) > m) m=i;
    };
    m+=3;

    ts=(char *)sf_malloc(m+24);

    printf("\tDN: [%s]\n", lgba_objectdn());
	for(vp = values; *vp; vp++) {
		p=lgba_map_friendly(*vp);
		printf("\t%s", p);
		i=m-strlen(p);
		printf(":");
		while(i-- > 0) printf(" ");
		if(lgba_gettype(*vp) & TY_BINARY) {
			slt = getattr2(*vp);
			if(slt)
				sprintf(ts, "[<Binary, size=%d>]\n", slt->lens[0]);
			else
				sprintf(ts, "[<Binary value>]\n");
		} else {
			sprintf(ts, "[${%s+], [}]\n", *vp);
		};
		printf("%s", getformat(ts));
    };
}

	return LGBA_NOERR;
};

int cmd_class(slist *sl) {
	int i;
	if(!lgba_objectname()) {
	fprintf(stderr,"-ERR: Object not loaded. Use load <object>.\n");
		return lgba_errno=LGBA_COMMAND; }
	if(sl->count == 0) {
	fprintf(stderr, "$INFO: Usage: class <objectclass> ...\n");
		return lgba_errno=LGBA_COMMAND; }
	lgba_errno=0;
	for(i=0; i<sl->count; i++)
		if(addclass(sl->list[i])) {
			fprintf(stderr, "-ERR: Add class '%s' failed: %s.\n", sl->list[i], lgba_err2string(lgba_errno));
			lgba_errno=LGBA_COMMAND;
		} else
	fprintf(stderr, "+OK: Class added.\n");
	return lgba_errno;
};

int cmd_rm(slist *sl) {
	int i;
	if(sl->count == 0 && !lgba_objectdn()) {
		fprintf(stderr, "$INFO: Usage: rm <object> ...\n");
		return lgba_errno=LGBA_COMMAND;
	} else if(sl->count == 0) {
		if(rmobject(NULL))
			lgba_perror("-ERR: Remove failed");
		else
			fprintf(stderr, "+OK: Object removed.\n");
	}
	for(i=0; i<sl->count; i++) {
		if(rmobject(sl->list[i]))
			fprintf(stderr, "-ERR: Remove '%s' failed: %s.\n", sl->list[i], lgba_err2string(lgba_errno));
		else
			fprintf(stderr, "+OK: Object removed.\n");
	}
	return LGBA_NOERR;
};

int
cmd_arm(slist *sl) {
	int i;
	int ret;

	if(!lgba_objectdn()) {
		fprintf(stderr, "-ERR: Object not loaded. Use load <object>.\n");
		return lgba_errno=LGBA_COMMAND;
	};
	if(sl->count == 0) {
		fprintf(stderr, "$INFO: Usage: arm <attribute>[=<value>] ...\n");
		return lgba_errno=LGBA_COMMAND;
	};

	ret=0;
	for(i=0; i<sl->count; i++) {
		if((ret |= rm_complex(sl->list[i], sl->list[i][sl->lens[i] + 1])))
			lgba_perror("-ERR");
		else
			fprintf(stderr, "+OK: removed.\n");
	}

	return ret;
};

int cmd_crm(slist *sl) {
	int i;
	if(!lgba_objectdn()) {
	fprintf(stderr,"-ERR: Object not loaded. Use load <object>.\n");
		return lgba_errno=LGBA_COMMAND; }
	if(sl->count == 0) {
	fprintf(stderr, "$INFO: Usage: crm <objectclass> ...\n");
		return lgba_errno=LGBA_COMMAND; }
	lgba_errno=0;
	for(i=0; i<sl->count; i++)
		if(rmclass(sl->list[i])) {
			fprintf(stderr, "-ERR: Remove class '%s' failed: %s.\n", sl->list[i], lgba_err2string(lgba_errno));
			lgba_errno=LGBA_COMMAND;
		} else
	fprintf(stderr, "+OK: Class removed.\n");
	return lgba_errno;
};

int cmd_set(slist *sl) {
	if(!lgba_objectdn()) {
		fprintf(stderr, "-ERR: Object not loaded. Use load <object>.\n");
		return lgba_errno=LGBA_COMMAND;
	}
	if(sl->count == 0) {
		fprintf(stderr,
	"$INFO: Usage: set <attribute> <value1> [ <value2> ... ]\n");
		return lgba_errno=LGBA_COMMAND;
	}
	if(setattr2(sl->list[0], mkarray(sl, 1)))
		fprintf(stderr, "-ERR: Set '%s' failed: %s\n", sl->list[0], lgba_err2string(lgba_errno));
	else
		fprintf(stderr, "+OK: Set '%s'.\n", sl->list[0]);
	return LGBA_NOERR;
};

int cmd_add(slist *sl) {
		if(!lgba_objectdn()) {
			fprintf(stderr, "-ERR: Object not loaded. Use load <object>.\n");
			return lgba_errno=LGBA_COMMAND;
		}
		if(sl->count == 0) {
			fprintf(stderr,
		"$INFO: Usage: add <attribute> <value1> [ <value2> ... ]\n");
			return lgba_errno=LGBA_COMMAND;
		};
		if(addattr2(sl->list[0], mkarray(sl, 1)))
			fprintf(stderr, "-ERR: Add '%s' failed: %s\n", sl->list[0], lgba_err2string(lgba_errno));
		else
			fprintf(stderr, "+OK: Add '%s'.\n", sl->list[0]);
		return LGBA_NOERR;
};

int cmd_ov(slist *sl) {
	long arg;
	if(sl->count == 0) {
		fprintf(stderr, "$INFO: Usage: override {1/0}\n");
		fprintf(stderr, "This statement used to control the 'override; functionality used in format().\n");
		return LGBA_NOERR;
	};
	arg = atoi(sl->list[0]);
	fprintf(stderr, "+OK: Set: %d.\n", lgba_ctl(LCTL_SET_OVERRIDEMODE, &arg));
	return LGBA_NOERR;
};

int cmd_cec(slist *sl) {
	frozen=0;
	return lgba_errno=LGBA_NOERR;
};

int cmd_fec(slist *sl) {
	fcode=lgba_errno;
	frozen=1;
	return 0;
};

int
cmd_file(slist *sl) {
	if(!sl->count && getuid() && geteuid()) {
		fprintf(stderr, "-ERR: Files must be named explicitly in non-root mode.\n");
		return lgba_errno=LGBA_PERM;
	};
	if(makefiles(sarray(sl, 0)))
		fprintf(stderr, "-ERR: No files defined.\n");
	return LGBA_NOERR;
};

int
cmd_vp(slist *sl) {
	int n;

	if(!lgba_objectdn()) {
		fprintf(stderr, "-ERR: Object not loaded. Use load <object>.\n");
		return lgba_errno=LGBA_COMMAND;
	};

	if(!sl->count) {
		lgba_errno=lgba_verifypwd(getpass("Verify password:"));
		if(lgba_errno == LGBA_NOERR)
			fprintf(stderr, "+OK: Password match\n");
		else
			fprintf(stderr, "-ERR: Password incorrect\n");
		return lgba_errno;
	};

	for(n=0; n < sl->count; n++) {
		lgba_errno=lgba_verifypwd(sl->list[n]);
		if(lgba_errno == LGBA_NOERR)
			fprintf(stderr, "+OK: Password match\n");
		else
			fprintf(stderr, "-ERR: Password incorrect\n");
	};

	return lgba_errno;
};

int
cmd_exit(slist *sl) {
	if(sl->count)
		return atoi(sl->list[1]);
	exit(frozen?fcode:lgba_errno);
};


#ifdef	HAVE_LIBREADLINE
char *objects_generator(char *text, int state) {
	char *oa=lgba_ObjectAttribute();
	static LDAPMessage *res;
	static LDAPMessage *en;
	static char **values=NULL;
	static char **vp=NULL;
	
	if(!state) {
		char *atl[] = { oa, NULL };
		char *flt=(char *)sf_malloc(strlen(text) + strlen(oa) + 4);
		sprintf(flt, "%s=%s*", oa, text);
		if(ldap_search_s(ld, lgba_current_dn(), LDAP_SCOPE_ONELEVEL, flt,
			atl, 0, &res)) {
			free(flt); return NULL;
		};
		free(flt);

		en=NULL;
		values=vp=NULL;
	};

	if(vp) {
		if(*vp) return *vp++;
		else { free(values); values=vp=NULL; };
	};

ogs:

	if(!en) {
		if(!(en=ldap_first_entry(ld, res)))
			{ ldap_msgfree(res); return NULL; };
	} else {
		if(!(en=ldap_next_entry(ld, en)))
			{ ldap_msgfree(res); return NULL; };
	};

	vp=values=ldap_get_values(ld, en, oa);

	if(!vp) goto ogs;
	if(!*vp) { free(values); goto ogs; }

	return *vp++;
};


char *attributes_generator(char *text, int state) {
	static char **vp=NULL;
	static int len;

	if(!state) {
		vp=lgba_objectattributes();
		if(!vp) return NULL;
		len=strlen(text);
	};

	while(*vp && strncasecmp(*vp, text, len)) vp++;

	if(*vp) return sf_strdup(*vp++);

	return NULL;
};
#endif
