#ifndef __LGBA_H__
#define __LGBA_H__

#include <strfunc.h>

#ifdef __cplusplus
extern "C" {
#endif

/* ============================================== */
/* ============ LGBA API definitions ============ */
/* ============================================== */

int lgba_init(char *ConfigFile);	/* Read config and initialize */
int lgba_open();			/* Open LDAP connection */
int lgba_close();			/* Close LDAP connection */

int lgba_chnode(const char *);	/* Change node */
char const *lgba_current_node();	/* Get node name */
slist *lgba_nodes();		/* Get nodes list */

int lgba_cd(const char *);	/* Change current DN (".." to level up) */
char *lgba_current_dn();	/* Get current DN */

int lgba_setuser(const char *); /* Change current user */

int lgba_ctl(int Flag, long *arg);	/* Set/Get internal states */

/*
** Flags for lgba_ctl are:
*/
#define	LCTL_GET_HTMLMODE	1	/* Default: 0 */
#define	LCTL_SET_HTMLMODE	2
#define	LCTL_GET_OVERRIDEMODE	3	/* Default: 1 */
#define	LCTL_SET_OVERRIDEMODE	4
#define	LCTL_GET_USE_LOCALE	5	/* Default: 1 */
#define	LCTL_SET_USE_LOCALE	6
#define	LCTL_GET_SIZELIMIT	7	/* N entries to ask from LDAP server */
#define	LCTL_SET_SIZELIMIT	8
#define	LCTL_GET_LOADFILTER	9	/* Restrict loading by filter */
#define	LCTL_SET_LOADFILTER	10

/****************************/
/*** Lower-level functons ***/
/****************************/

int load(const char *objectname);
char *lgba_objectname();
char *lgba_objectdn();
char **lgba_objectclass();
char **lgba_objectattributes();

int lgba_save(int free);  /* Save changes [&destroy] the current object */
void lgba_free();	/* Destroy current object, ignoring changes */
char **lgba_explode_dn(char *);
void lgba_setattrlist(char **); /* for loading only the necessary values */
int lgba_verifypwd(const char *password); /* compare cleartext password */
int lgba_exists(const char *); /* is object exists? */

/********************/
/*** Start search ***/
/********************/

/*
 * base:	base DN (may be NULL or "")
 * filter:	a standart LDAP filter, like (&(objectclass=*)(cn=*fred*))
 * attrlist:	an optional NULL-terminated attributes list, for minimizing
 *		the overhead by getting over TCP/IP only the necessary attrs
 * sortby:	an optional attribute by which we will sort the entries
 * cmp:		the optional comparison function
 * counted:	(optional) invoke this function with the quantity of entries
 *		right after the search completion. function must be defined
 *		as int counted(int);
 * func:	invoke this function for every entry
 * flags:
 *	0: default one level search
 *  1: search subtree instead of default "one level"
 *  2: replay the last search (base, filter and attrlist are ignored)
 *
 */

int lgba_search(char *optBase, char *optFilter, char **optAttrlist, char *optSortby, int (*optCmp)(const char *a, const char *b), int (*optCounted)(int), int (*optFunc)(int), int flags);

/*
** Useful alternative to lgba_search(). sortby is NULL-terminated array
*/

struct	SortBy {
	const char *attr; // Attribute name 
	const char *dflt; // Default value, if attribute is empty
	int (*cmp)(const char *a, const char *b);
};

int lgba_search2(char *base, char *filter, char **attrlist, struct SortBy **sortby, int (*counted)(int), int (*func)(int), int flags);

/***************************************************/
/*** Browse through the all available attributes ***/
/***************************************************/

char *lgba_first_attribute();
char *lgba_next_attribute();

/*******************************************************/
/*** Get an attribute (list) from the current object ***/
/*******************************************************/

char *getattr(const char *attribute);

/*
** You might want to get the binary values using the ->lens[] member of
** the slist* structure.
*/
slist *getattr2(const char *attribute);

/**********************************************************/
/*** Parse fmtstring against ${}, $() and $`' variables ***/
/**********************************************************/

char *getformat(char *fmtstring);	// leak-safe
int format(char *fmtstring, int fd);

/********************/
/*** Type control ***/
/********************/

/*
** Possible types:
*/
#define TY_SINGLE	1
#define TY_BINARY	2
#define TY_NUMERIC	4
#define TY_PASSWD	8
#define TY_CRYPTPW	16
#define TY_TEXT		32
#define TY_MAP		64

int lgba_gettype(const char *attribute);

/*
** Type-specific things:
** For the TY_MAP:	Filename of the map
** For the TY_BINARY:	The MIME-type
*/
void *lgba_gettypespec(const char *attribute);

/**********************/
/*** Access control ***/
/**********************/

int canIcreate(const char *attribute);
int canIread(const char *attribute);
int canIwrite(const char *attribute);
int canIdelete(const char *attribute);

/*******************************/
/*** Special attribute names ***/
/*******************************/

char *getmodifierattributename();
char *getmodifiedattributename();
char *getcreatorattributename();
char *getcreatedattributename();

/* Get object attribute */
char *lgba_ObjectAttribute(); 	/* probably, will return "object" */

/***********************************************/
/*** Modify and delete attributes and object ***/
/***********************************************/

int create(const char *object, const char *cls);	/* class may be empty */
int setattr(const char *attribute, const char *value);
int setattr2(const char *attribute, char **values); /* frees **values!!! */
int addattr(const char *attribute, char *value);
int addattr2(const char *attribute, char **values); /* frees **values!!! */
int rmattr(const char *attribute, const char *optValue); /* remove attribute value */
int rmattr2(const char *attribute, char **optValues); /* frees **optValues!!! */
int rmobject(const char *obj);	/* Removes obj or current object if obj==NULL */
int addclass(const char *cls); /* add a class to an existing object */
int rmclass(const char *cls); /* remove a class from an existing object */

/********************************/
/*** Create a formatted files ***/
/********************************/

int makefiles(char **names_list);

char **lgba_get_classnames();
char **lgba_getwritelist();

char *lgba_setlang(char *);	/* set default language */
char *lgba_map_friendly(const char *); /* perform a friendly mapping */

void lgba_addfriendly(char *, char *);	/* add friendly mapping */
void lgba_addlanguage(const char *, char **, char **);

/*** other ***/

/* Read an attribute-value file */
int lgba_readavfile(const char *filename, slist **a, slist **v);

extern int lgba_errno;
void lgba_perror(const char *);
char *lgba_err2string(int err); /* self-descriptive... */

#define LGBA_NOERR	0
#define LGBA_UNKNOWN	1
#define LGBA_CONF	2
#define LGBA_CONN	3
#define LGBA_PERM	4
#define LGBA_LDAP	5
#define LGBA_COMMAND	6
#define LGBA_SETLANG	7
#define LGBA_ACCESS	8
#define LGBA_NOTFOUND	9
#define LGBA_MODIFY	10
#define LGBA_BIND	11
#define LGBA_VALUE	12
#define LGBA_NOCLASS	13
#define LGBA_MANYVALS_DEPR	14
#define LGBA_CREATE		15
#define LGBA_NODENF		16
#define LGBA_ALREADY	17
#define LGBA_ALEX		18
#define LGBA_RSHORT	19
#define LGBA_NOATTR	20
#define LGBA_TOOMANY	21

#ifdef __cplusplus
}
#endif

#endif
