/*
 * ====================================================================
 * Copyright (c) 1995-2000 Lyonel VINCENT and Jeff Morrow.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mod_ldap.c: LDAP-based access checking for Apache
 *
 * Lyonel VINCENT (vincent@trotek.ec-lyon.fr) and Jeff Morrow
 *   (jmorrow@alum.mit.edu)
 * changes by Sam Alcoff (alcoff@survival.net) to compile with
 *   Apache 1.3
 * changes by Rick Perry (perry@ece.vill.edu): bugs fixes and code
 *   cleanups.
 * changes by Lev Walkin (vlm@spelio.net.ru): simplifying for LGBA
 * 
 * Other contributors:
 *   Markku.Jarvinen@tpo.fi
 *   Soren.Roug@EEA.eu.int
 *   Dan Rich (drich@solipsa.com)
 *
 * Version's 1.7 features were extremely reduced by vlm.
 * Version 1.6 (July 1999) -- a few feature additions and bug fixes
 *   (including a seg fault fix) by Markku Jarvinen and Soren Roug.
 *   Also compiles warning-free on Linux/gcc.
 * Version 1.5 (Jan 1999) -- adds several new features and merges
 *   the functionality of mod_ldap 1.4 with mod_auth_ldap 1.1.
 *
 * based on the mod_auth_ldap.c module written by
 * Norman RICHARDS (orb@cs.utexas.edu)
 *
 * =============================================
 *
 * For more information on LDAP, see the LDAPWorld web 
 * page at:  http://www.critical-angle.com/ldapworld/
 *
 * =============================================
 *
 * The goal of this module is to allow access checking 
 * through user information stored in an LDAP directory. 
 *
 * This module works very similarly to other access checking
 * modules.
 * 
 * To use the module, you need to add this file to your 
 * apache src directory.  Add the following line to your
 * Configuration file:
 *
 *     Module ldap_module     mod_ldap.o
 *
 * Alternatively, if you have placed mod_ldap.c in the directory
 * src/modules/extra, you may add the following flag to your
 * configure directive:
 *
 *     --activate-module=src/modules/extra/mod_ldap.o
 *
 * In either case, you will also need to add the appropriate
 * locations for your ldap/lber libraries and includes to 
 * EXTRA_LIBS and EXTRA_INCLUDES.
 *
 * =============================================
 *
 * The following directives can be used in your access.conf
 * file in the same way as you would configure the standard
 * authentication modules.
 *
 * LDAPAuth <flag>
 *
 * LDAPServer <LDAP URL>
 *
 *   This sets the LDAP server to be used.  The ldap url 
 *   is of the form "ldap://yourhost.com:portnumber/"
 *
 * LDAPBindName <Distinguished Name>
 * LDAPBindPass <password>
 *
 *   If the module needs to authenticate itself, enter the user 
 *   name and password here.
 *
 * LDAPBase <Base DN>
 *
 *   Specifies the base DN for all LDAP searches.
 *
 * LDAPuseridAttr <attribute name>
 *
 *   Default for LDAPuseridAttr is "object" (LGBA convention).
 *
 * LDAPFilter <LDAP search filter>
 *
 *   The search filter is in LDAP standard format as specified by
 *   RFC1960
 *
 * The require directives are as follows:
 *
 * require valid-user
 *
 *   Allows any valid LDAP user.
 *
 * =============================================
 *
 * example:
 *
 * AuthType Basic
 * AuthName "Vens Emission system"
 * LDAPAuth On
 * LDAPServer "ldap://ns.vens.ru/"
 * LDAPBase "domain=vens.ru, o=Vens"
 * LDAPBindName "cn=root, o=Vens"
 * LDAPBindPass "nskdutbw"
 * LDAPuseridAttr  object
 * LDAPFilter "(|(object=vlm)(object=slayer))"
 * <limit GET POST>
 * require valid-user
 * </limit>
 *
 */

#include "ap_compat.h"

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#include <lber.h>
#include <ldap.h>

module ldap_module;


/* --------------------------------- */

typedef struct  {
  char *host;
  int  port;
  
  char *base;
  char *bindname;
  char *bindpass;
  char *filter;
  int ldap_auth;
  
  char *userid_attr;
  
} ldap_config_struct;


void *create_ldap_dir_config(pool *p, char *d) {
  ldap_config_struct *conf;
  
  conf = (ldap_config_struct *) ap_pcalloc(p, sizeof(ldap_config_struct));
  
  conf->host       = "localhost";
  conf->port       = LDAP_PORT;
  
  conf->base       = NULL;
  conf->bindname   = NULL;
  conf->bindpass   = NULL;

  conf->filter     = NULL;
  
  conf->userid_attr		= "object";
  
  conf->ldap_auth  = 0;
  
  return conf;
}

/* --------------------------------- */

const char *set_ldap_auth(cmd_parms *cmd, ldap_config_struct *conf, int arg)
{
  conf->ldap_auth = arg;
  return NULL;
}

const char *set_ldap_base(cmd_parms *cmd, ldap_config_struct *conf, char *base)
{
  conf->base=ap_pstrdup(cmd->pool, base);
  return NULL;
}

const char *set_ldap_useridattr(cmd_parms *cmd, ldap_config_struct *conf, char *attr)
{
  conf->userid_attr=ap_pstrdup(cmd->pool, attr);
  return NULL;
}

const char *set_ldap_server(cmd_parms *cmd, ldap_config_struct *conf, char *url) {
  LDAPURLDesc *ldap_url;
  
  if (!ldap_is_ldap_url(url)) {
    return "server is not a properly formatted LDAP url";
  }
  
  if (ldap_url_parse(url,&ldap_url)!=0) {
    return "cannot parse LDAP url";
  }
  
  conf->host=ap_pstrdup(cmd->pool, ldap_url->lud_host);
  if (ldap_url->lud_port!=0)  
    conf->port=ldap_url->lud_port; 
  
  ldap_free_urldesc(ldap_url);
  
  return NULL;
}

const char *set_ldap_bindname(cmd_parms *cmd, ldap_config_struct *conf, char *name) {
  conf->bindname = ap_pstrdup(cmd->pool,name);  
  return NULL;
}

const char *set_ldap_filter(cmd_parms *cmd, ldap_config_struct *conf, char *filter) {
  conf->filter = ap_pstrdup(cmd->pool,filter);  
  return NULL;
}

const char *set_ldap_bindpass(cmd_parms *cmd, ldap_config_struct *conf, char *pass) {
  conf->bindpass = ap_pstrdup(cmd->pool,pass);  
  return NULL;
}

command_rec ldap_auth_cmds[] = {
  { "LDAPAuth",     set_ldap_auth,     NULL, OR_AUTHCFG, FLAG,    "Activate LDAP auth" },
  { "LDAPServer",     set_ldap_server,     NULL, OR_AUTHCFG, TAKE1,    "LDAP URL" },
  { "LDAPBase",     set_ldap_base,     NULL, OR_AUTHCFG, TAKE1,    "LDAP search base" },
  { "LDAPuseridAttr",     set_ldap_useridattr,     NULL, OR_AUTHCFG, TAKE1,    "LDAP user id attribute name" },
  { "LDAPFilter",     set_ldap_filter,     NULL, OR_AUTHCFG, TAKE1,    "LDAP filter" },
  { "LDAPBindName",   set_ldap_bindname,   NULL, OR_AUTHCFG, RAW_ARGS, NULL },
  { "LDAPBindPass",   set_ldap_bindpass,   NULL, OR_AUTHCFG, RAW_ARGS, NULL },
  {NULL}
};

/* --------------------------------- */
int ldap_check(request_rec *r, ldap_config_struct *conf, char *uname, char *pwd) {
  LDAP *ld;
  char *flt;
  char *avl[]={ conf->userid_attr, NULL };
  LDAPMessage *res, *rq;
  
  if (uname == NULL)
	return AUTH_REQUIRED;
  if(!pwd || !*pwd)
	return 0;

  if(strspn(uname, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-") != strlen(uname)) return 0;

  if(!(ld=ldap_init(conf->host, conf->port)))
	return 0;

  flt=(char *)malloc(strlen(uname) + strlen(conf->base) + 16);
  if(!flt)
	return 0;

  sprintf(flt, "%s=%s, %s", conf->userid_attr, uname, conf->base);
  if(ldap_simple_bind_s(ld, flt, pwd) == LDAP_SUCCESS) {
	int k;

	if(!conf->filter) {
		free(flt);
		ldap_unbind(ld);
		return 1;
	};

  	if(k=ldap_search_s(ld, flt, LDAP_SCOPE_BASE, conf->filter,avl,0,&res)){
		free(flt);
		ldap_unbind(ld); return 0;
	};

	free(flt);
	k=ldap_count_entries(ld, res);
	ldap_msgfree(res);
	if(k==1) { ldap_unbind(ld); return 1; };
  };

  if(ldap_simple_bind_s(ld, conf->bindname,
	conf->bindname?conf->bindpass:NULL)) {
	free(flt); ldap_unbind(ld); return 0;
  };

  flt=(char *)realloc(flt, strlen(uname) +
	strlen(conf->userid_attr) +
	((conf->filter)?(strlen(conf->filter)):0) + 24);

	if(!flt)
		return 0;

  if(conf->filter)
	sprintf(flt, "(&(%s=%s)%s)", conf->userid_attr, uname, conf->filter);
  else
	sprintf(flt, "(%s=%s)", conf->userid_attr, uname);

  if(ldap_search_s(ld, conf->base, LDAP_SCOPE_ONELEVEL, flt, avl, 0, &res)) {
	free(flt);
	ldap_unbind(ld);
	return 0;
  };

  free(flt);

  rq=ldap_first_entry(ld, res);
  while(rq) {
	char *dn=ldap_get_dn(ld, rq);
  	if(dn && ldap_simple_bind_s(ld, dn, pwd) == 0) {
  		free(dn);
  		ldap_msgfree(res);
		ldap_unbind(ld);
		return 1;
	};
	free(dn);
	rq=ldap_next_entry(ld, rq);
  };

  ldap_msgfree(res); ldap_unbind(ld); return 0;
};

int ldap_authenticate (request_rec *r)
{
  ldap_config_struct *conf;
  conn_rec *c = r->connection;
  const char *sent_pw;
  int res;
  
  conf = (ldap_config_struct *) 
    ap_get_module_config(r->per_dir_config, &ldap_module);
  
  if(!conf->ldap_auth) return DECLINED;

  if (ap_get_basic_auth_pw(r, &sent_pw) || !*sent_pw) return AUTH_REQUIRED;

  if(!ldap_check(r, conf, r->connection->user, (char *)sent_pw))
	return AUTH_REQUIRED;

  return OK;
};

int ldap_check_auth (request_rec *r) {
  ldap_config_struct *conf;
  array_header *reqs_arr;
  require_line *reqs;
  char *userdn;
  int i;
  int active = 0;
  char *t, *w;

  /* DEBUGGING */
  /* ap_log_reason("Entering ldap_check_auth",r->uri,r); */
  
  conf = (ldap_config_struct *) 
    ap_get_module_config(r->per_dir_config, &ldap_module);
  reqs_arr = (array_header *)(ap_requires (r));
  reqs = reqs_arr ? (require_line *)reqs_arr->elts : NULL;
  
  /* must have require field(s) */
  if (!reqs_arr || !r->connection->user) {
    return DECLINED;
  }
  
  /* Changed by Soren Roug */

  /* DEBUGGING */
  /* ap_log_reason(ap_pstrcat(r->pool, "reqs_arr->nelts = ", num2str(reqs_arr->nelts), NULL),r->uri,r); */

  for(i=0; (i < reqs_arr->nelts) ; i++) {
    if (! (reqs[i].method_mask & (1 << (r->method_number)))) continue;
    
    t = reqs[i].requirement;
    w = ap_getword(r->pool, (const char **)(&t), ' ');
    
    /* DEBUGGING */
    /* ap_log_reason(ap_pstrcat(r->pool, "w = ", w, NULL),r->uri,r); */
    
    if (strcmp(w,"valid-user")==0)
	{
	  if(!conf->ldap_auth)
	    return DECLINED;
	  else
            {
	      active = 1;
	      return OK;
            }
	}
  }
    
  if(!active) return DECLINED;
  
  ap_log_reason(ap_pstrcat(r->pool,
			   "LDAP access denied for ",
			   r->connection->user,
			   NULL),
		r->uri,r);
  return AUTH_REQUIRED;
}

module ldap_module = {
  STANDARD_MODULE_STUFF,
  NULL,				/* initializer */
  create_ldap_dir_config,	/* dir config creater */
  NULL,				/* dir merger - default=override */
  NULL,				/* server config */
  NULL,				/* merge server config */
  ldap_auth_cmds,		/* command table */
  NULL,				/* handlers */
  NULL,				/* filename translation */
  ldap_authenticate,		/* check_user_id */
  ldap_check_auth,		/* check auth */
  NULL,				/* check access */
  NULL,				/* type_checker */
  NULL,				/* pre-run fixups */
  NULL				/* logger */
};
