/*-
 * Copyright 1998 Juniper Networks, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/time.h>
#include <pwd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <lgba.h>

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#include <security/pam_modules.h>

#include "pam_mod_misc.h"

#define PASSWORD_PROMPT	"Password:"

/*
 * authentication management
 */

#define	OPT_CONF	"conf="
#define	OPT_CONF_S	(sizeof(OPT_CONF)-1)
#define	OPT_TMPL	"template_user="
#define	OPT_TMPL_S	(sizeof(OPT_TMPL)-1)
#define	OPT_NODE	"node="
#define	OPT_NODE_S	(sizeof(OPT_NODE)-1)

char *l_conf=NULL;
char *l_tmpl=NULL; /* "lgba"; */
char *l_node=NULL;

PAM_EXTERN int
pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
    const char **argv)
{
	int retval;
	const char *user;
	const char *password;
	int options;
	int i;

	options = 0;
	for (i = 0;  i < argc;  i++) {
		pam_std_option(&options, argv[i]);
		if( strncmp(argv[i], OPT_CONF, OPT_CONF_S) == 0 )
			l_conf=(char *)argv[i] + OPT_CONF_S;
		else if( strncmp(argv[i], OPT_TMPL, OPT_TMPL_S) == 0 )
			l_tmpl=(char *)argv[i] + OPT_TMPL_S;
		else if( strncmp(argv[i], OPT_NODE, OPT_NODE_S) == 0 )
			l_node=(char *)argv[i] + OPT_NODE_S;
	};

	if ((retval = pam_get_user(pamh, &user, NULL)) != PAM_SUCCESS)
		return retval;

	if ((retval = pam_get_pass(pamh, &password, PASSWORD_PROMPT,
	    options)) != PAM_SUCCESS)
		return retval;

	if(lgba_init(l_conf) || lgba_open())
		return PAM_SERVICE_ERR;

	if(l_node && lgba_chnode(l_node)) {
		lgba_close();
		return PAM_AUTHINFO_UNAVAIL;
	};

	{
		char *avl[] = {
			lgba_ObjectAttribute(),
			"objectclass", "cn",
			"uidnumber", "gidnumber",
			NULL };
		lgba_setattrlist(avl);
	};

	if(load(user)) {

#ifndef	PREPEND
		lgba_close();
		return PAM_AUTH_ERR;
#else

		char *newuser;

		newuser = (char *)malloc(strlen(user) + 2);
		if(!newuser) {
			lgba_close();
			return PAM_AUTH_ERR;
		};
		*newuser='*';
		strcpy(newuser+1, user);

		if(load(newuser)) {
			free(newuser);
			lgba_close();
			return PAM_AUTH_ERR;
			/* Insecure way of returning authentication error.
			return PAM_USER_UNKNOWN;
			*/
		};
		free(newuser);

		user = lgba_objectname();

		pam_set_item(pamh, PAM_USER, user);

#endif

	};

	if(lgba_verifypwd(password)) {
		lgba_close();
		return PAM_AUTH_ERR;
	};

	if(l_tmpl && getpwnam(user) == NULL) {
		pam_set_item(pamh, PAM_USER, l_tmpl);
	};

	if(getattr2("uidnumber"))
		pam_putenv(pamh, getformat("uidnumber=${uidnumber}"));
	if(getattr2("gidnumber"))
		pam_putenv(pamh, getformat("gidnumber=${gidnumber}"));

	lgba_close();

	return PAM_SUCCESS;
}

PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	return PAM_SUCCESS;
}

PAM_MODULE_ENTRY("pam_lgba");
